from enrichsdk.app.utils import EnrichAppConfig

class APPNAMEConfig(EnrichAppConfig):
    name = 'APPNAME'
    verbose_name = "APPDESC"
    description = f"APPDESC"
    status = "stable"
    enable = True
    filename = __file__
    multiple = False
    composition = True



#!/usr/bin/env python3

# Core
import os
import sys
import pip
import base64
import subprocess
import json
import imp
import yaml
import traceback
import pkg_resources
import logging, traceback
import pyfiglet

# thirdparty
import click
from pythonjsonlogger import jsonlogger

# self
import enrichsdk
from enrichsdk import (
    package,
    datasets,
    api as sdkapi,
    lib,
    featurestore as fslib,
    utils,
)


@click.group()
def process():
    """
    Bootstrap/test/install Enrich modules and access server

    \b
    Getting started:
       version:  Version of this sdk
       start:    First time instructions
       env:      Setup the environment context

    \b
    Development:
       bootstrap:  Bootstrap modules including transforms
       test:      Test transforms, manage datasets

    \b
    Server:
       api:       Access the server API

    \b
    Utils:
       sample:    Sample data for sharing

    \b
    Helpers:
       show-log:  Pretty print log output
    """
    pass


@click.command("version", hidden=True)
def _version():
    """
    Show version
    """
    print(enrichsdk.__version__)


@click.command("start", hidden=True)
@click.option("--commands/--no-commands", default=False, help="Show commands")
def _start(commands):
    """
    Instructions to get started
    """

    result = pyfiglet.figlet_format("Enrich SDK", font="slant")
    print(result)

    home = os.environ["HOME"]
    if "ENRICH_ROOT" in os.environ:
        enrich_root = os.environ["ENRICH_ROOT"]
    else:
        enrich_root = "$ENRICH_ROOT"

    if "VIRTUAL_ENV" in os.environ:
        venv_root = os.environ["VIRTUAL_ENV"]
        venv_name = os.path.basename(venv_root)
        install_settings = [
            'echo "source {}/bin/activate" > $ENRICH_ROOT/env.sh'.format(venv_root),
            'echo "#workon {}" >> $ENRICH_ROOT/env.sh'.format(venv_name),
        ]
    elif "CONDA_PREFIX" in os.environ:
        venv_root = os.environ["CONDA_PREFIX"]
        venv_name = os.path.basename(venv_root)
        install_settings = [
            'echo "conda activate {}" >> $ENRICH_ROOT/env.sh'.format(venv_name),
        ]
    else:
        print("Please use virtualenvwrapper or conda to install sdk")
        return

    steps = [
        {
            "description": "Understand Enrich and SDK",
        },
        {
            "description": "set ENRICH_ROOT and populate",
            "cmd": [
                "export ENRICH_ROOT={}/work/enrich".format(home),
                "mkdir -p $ENRICH_ROOT",
                "enrichpkg env populate",
            ],
        },
        {
            "description": "Check and update siteconf and versionmap",
            "cmd": [
                "enrichpkg env check",
                "cat $ENRICH_ROOT/etc/siteconf.json",
            ],
        },
        {
            "description": "[OPTIONAL] For complex deployments, create an environment/context file",
            "cmd": "enrichpkg env sample-context > context.yaml",
        },
        {
            "description": "[OPTIONAl] Create a simple settings file",
            "cmd": [
                "# file to be sourced before you start working. use the appropriate",
                "# environment activation mechanism. Check the script generated by",
                "# populate command",
                "$ cat $ENRICH_ROOT/env.sh"
                "",
                "If the file doesnt exist or doesnt look right, use the following"
            ]
            + install_settings
            + [
                'echo "export ENRICH_ROOT={}" >> $ENRICH_ROOT/env.sh'.format(
                    enrich_root
                )
            ],
        },
        {
            "description": "Change dir to $ENRICH_CUSTOMERS",
            "cmd": "cd $ENRICH_ROOT/customers",
        },
        {
            "cmd": [
                "#To handle python paths etc. avoid spaces and hyphen in the names",
                "git clone https://github.com/pingali/enrich-acme.git acme",
            ],
            "description": "GIT checkout a code repository",
        },
        {
            "cmd": "cd $ENRICH_ROOT/customers/acme",
            "description": "Change to checked out repository",
        },
        {
            "cmd": ["enrichpkg bootstrap -c repo -p ."],
            "description": "Bootstrap the repo if not already done",
        },
        {
            "cmd": [
                "# Typically usecases have first letter in capital",
                "enrichpkg bootstrap -c usecase -p Marketing",
            ],
            "description": "Create usecase, say Marketing",
        },
        {"cmd": ["cd Marketing"], "description": "Change to usecase"},
        {
            "description": "Bootstrap a transform. Transform is a python module. It could be a single file, directory or a full package",
            "cmd": [
                "# Will create an example hello world script",
                "enrichpkg bootstrap -c transform-helloworld -p transforms/helloworld.py",
                "",
                "# Will create a more simple python script that is more complex than helloworld",
                "enrichpkg bootstrap -c transform-simple -p transforms/simplefile.py",
                "",
                "# Will create a python module (a full directory)",
                "enrichpkg bootstrap -c transform-simple -p transforms/simpledir",
                "",
                "# Will create a python package. Dont use this unless you are sure",
                "enrichpkg bootstrap -c transform-package -p transforms/simplepackage",
            ],
        },
        {
            "cmd": [
                "# Add any requirements and install them",
                "vi ../requirements.txt",
                "pip install -r ../requirements.txt",
            ],
            "description": "Install repo-specific requirements",
        },
        {
            "cmd": [
                "enrichpkg test transform transforms/helloworld.py",
                "# To capture all debug logs",
                "enrichpkg test transform --capture transforms/helloworld.py",
            ],
            "description": "Test the transform",
        },
        {
            "cmd": [
                "enrichpkg bootstrap -c pipeline-conf -p pipelines/conf/helloworld.py"
            ],
            "description": "Bootstrap a pipeline",
        },
        {
            "cmd": ["enrichpkg test conf --capture pipelines/conf/helloworld.py"],
            "description": "Check pipeline",
        },
        {
            "cmd": [
                "# prefect is the default workflow engine",
                "enrichpkg bootstrap -c prefectjob -p workflows/prefect/daily.py",
            ],
            "description": "Bootstrap a Prefect workflow",
        },
        {"cmd": ["workflows/prefect/daily.py"], "description": "Check workflow"},
    ]

    if not commands:
        print("Note: Use --commands to see how to implement these steps\n")

    for i, s in enumerate(steps):
        print("[{:2}] {}".format(i + 1, s["description"]))
        if not commands:
            continue

        print("")
        if "cmd" in s:
            if isinstance(s["cmd"], str):
                print("    $ {}".format(s["cmd"]))
            else:
                for c in s["cmd"]:
                    print("    $ {}".format(c))
        print("")


@click.command("bootstrap", hidden=True)
@click.option(
    "--component",
    "-c",
    required=True,
    help="What should be bootstrapped",
    type=click.Choice(
        [
            "repo",
            "usecase",
            "transform-package",
            "transform-simple",
            "transform-helloworld",
            "transform-query",
            "asset",
            "dashboard",
            "task-lib",
            "task-conf",
            "pipeline-conf",
            "sparkjob",
            #'airflow',
            "prefectjob",
            "rscript",
            "pyscript",
        ]
    ),
)
@click.option(
    "--path", "-p", required=True, help="Directory where the files should be stored"
)
@click.option(
    "--template",
    "-t",
    required=False,
    default=None,
    help="Template to override the defaults",
)
@click.option("--context", default=None, help="Environment file")
def bootstrap_(component, path, template, context):
    """
    Bootstrap a fresh module
    """

    """
    Modules currently supported include transform, model, and skin.

    For now only the transforms and models are testable locally
    without the need for a test server.

    In future we expect to more comprehensively support the
    development of all modules.

    A custom template can be provided as well.
    """

    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        raise
    context.set_env()

    if not package.checkenv(validate_only=True):
        print("Environment is incomplete")
        print("Try: enrichpkg env check")
        return

    package.bootstrap(component, path, template)


@click.command("show-log", hidden=True)
@click.argument("logfile")
def _show_log(logfile):
    """
    Pretty-print run log
    """

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    package.print_log(logfile)


@click.group("env", hidden=True)
@click.pass_context
def envsetup(ctx):
    """
    Setup the environment. Environment includes:\n
      \b
      (a) Workspace for enrich to work
      (b) Minimal configuration

    This can be specified using environment variables
    and/or 'context' file.

    The absolute minimal required environment variable
    is ENRICH_ROOT (path to enrich workspace e.g.,
    ~/enrich)
    """

    ctx.ensure_object(dict)


@envsetup.command("sample-context")
@click.option("--root", "-r", default=None, help="Root of enrich")
def _context_generate(root):
    """
    Generate sample context file
    """

    if root is not None:
        os.environ["ENRICH_ROOT"] = os.path.abspath(root)
    elif "ENRICH_ROOT" not in os.environ:
        os.environ["ENRICH_ROOT"] = os.path.expandvars("$HOME/enrich")

    context = lib.Context()
    context = context.asdict()

    if "siteconf" not in context:
        context["siteconf"] = os.path.join(context["ENRICH_ETC"], "siteconf.json")

    print("Please store this in context.yaml")
    print("Minimum required valid paths include: siteconf, versionmap, enrich_data")
    print("Example: enrichpkg api --context context.yaml ...")
    print("----")
    print(yaml.dump(context, indent=4))


@envsetup.command("sample-versionmap")
def _versionmap():
    """
    Generate sample versionmap
    """

    print(
        "Please store this in versionmap.json. Default is $ENRICH_ROOT/etc/versionmap.json"
    )
    print("Update context.yaml with this versionmap path")
    print("----")
    print(package.get_sample_versionmap())


@envsetup.command("sample-siteconf")
def _siteconf():
    """
    Generate sample siteconf
    """

    print(
        "Please store this in siteconf.json. Default is $ENRICH_ROOT/etc/siteconf.json"
    )
    print("Update context.yaml with this siteconf path")
    print("----")
    print(package.get_sample_siteconf())


@envsetup.command("populate")
@click.option("--context", default=None, help="Context file")
def _populate(context):
    """
    Populate the directories
    """
    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        print("Error! " + str(e))
        return

    context.set_env()

    package.populate(context=context.asdict())


@envsetup.command("check")
@click.option("--context", default=None, help="Context file")
def _checkenv(context):
    """
    Check environment
    """

    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        print("Error! " + str(e))
        return

    context.set_env()

    package.checkenv(context=context)


##########################################################
# Test command
##########################################################
@click.group("test", hidden=True)
@click.option("--context", default=None, help="Environment file")
@click.pass_context
def test(ctx, context):
    """
    Test transforms/pipelines etc.
    """

    # use the context file to set the environment
    try:
        context = lib.Context(context)
    except Exception as e:
        raise

    context.set_env()
    ctx.ensure_object(dict)
    ctx.obj["context"] = context.asdict()


@test.command("transform")
@click.argument("pkgdir")
@click.option("--capture/--no-capture", default=True, help="Capture output")
@click.pass_context
def test_pkg(ctx, pkgdir, capture):
    """
    Unit testing of a package module (transform)
    """
    context = ctx.obj["context"]

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    package.test_transform(pkgdir, capture, context)


@test.command("task-lib")
@click.argument("taskdir")
@click.option("--capture/--no-capture", default=True, help="Capture output")
@click.pass_context
def test_task(ctx, taskdir, capture):
    """
    Unit testing of a task library
    """

    context = ctx.obj["context"]

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    package.test_task(taskdir, capture, context)


@test.command("conf")
@click.argument("spec")
@click.option("--capture/--no-capture", default=True, help="Capture output")
@click.pass_context
def test_conf(ctx, spec, capture):
    """
    Minimal testing of pipeline/task configuration
    """

    context = ctx.obj["context"]

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    package.test_conf(spec, capture, context)


@click.group("data")
@click.argument("spec")
@click.option("--capture/--no-capture", default=True, help="Capture output")
@click.pass_context
def testdata(ctx, spec, capture):
    """
    Manage test data


    spec could be transform or a spec file
    """
    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    ctx.ensure_object(dict)
    ctx.obj["spec"] = spec
    ctx.obj["capture"] = capture


@testdata.command("list")
@click.pass_context
def _testdata_list(ctx):
    """
    List available test datasets
    """

    package.testdata_action(
        "list", ctx.obj["spec"], ctx.obj["capture"], ctx.obj["context"]
    )


@testdata.command("show")
@click.argument("dataset")
@click.pass_context
def _testdata_show(ctx, dataset):
    """
    Show the details of a given dataset
    """

    package.testdata_action(
        "show", ctx.obj["spec"], ctx.obj["capture"], ctx.obj["context"], name=dataset
    )


@testdata.command(
    "run",
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
@click.argument("dataset")
@click.argument("command")
@click.pass_context
def _testdata_action(ctx, dataset, command):
    """
    Generate action commands for a dataset over a specified range

    Pass additional parameters as key-value pairs as
    shown in show.
    """

    args = dict()
    for item in ctx.args:
        args.update([item.split("=")])

    package.testdata_action(
        "action",
        ctx.obj["spec"],
        ctx.obj["capture"],
        ctx.obj["context"],
        name=dataset,
        command=command,
        **args
    )


test.add_command(testdata)

#########################################
# API
#########################################
@click.group(hidden=True)
@click.option("--config", default=None, help="API Config File")
@click.pass_context
def api(ctx, config):
    """
    Access Enrich Server
    """

    if not package.checkenv(validate_only=True):
        print("Try: enrichpkg checkenv")
        raise Exception("Environment incomplete")

    alt = os.path.expandvars("$ENRICH_ETC/api.json")
    if config is None:
        config = alt

    # Write a default
    default = {"schema": "v1:api:enrich", "backends": []}
    if not os.path.exists(config):
        with open(config, "w") as fd:
            fd.write(json.dumps(default, indent=4))

    ctx.ensure_object(dict)
    ctx.obj["config"] = config


####################################
# Support for sampling data
####################################
@click.group("sample", hidden=True)
@click.pass_context
def sample(ctx):
    """
    Sample data from blockstores
    """
    pass


@sample.command("crawl")
@click.argument("root")
@click.option("--ext", default="json", help="Extension of files to look for")
@click.pass_context
def crawl(ctx, root, ext):
    """
    Crawl the root to find possible files

    Args:
       root: Path in blockstore with s3:// prefix
    """

    utils.sample_crawl(root, ext)


@sample.command("cat")
@click.argument("path")
@click.pass_context
def crawl(ctx, path):
    """
    Show the contents of path

    Args:
       path: Path to file in blockstore with s3:// prefix
    """

    utils.sample_cat(path)


####################################
# Configure access to server
####################################
@click.group()
@click.pass_context
def config(ctx):
    """
    Configure access to Enrich Server
    """
    pass


@config.command("init")
@click.pass_context
def _config_init(ctx):
    """
    Initialize the config file
    """
    default = {
        "schema": "v1:api:enrich",
        "backends": [],
        "services": [],
    }
    filename = ctx.obj["config"]
    with open(filename, "w") as fd:
        fd.write(json.dumps(default, indent=4))

    print("Initialized config")


@config.command("list")
@click.pass_context
def _config_list(ctx):
    """
    List available backends/services
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    print("Backends")
    print("-----")
    backends = config["backends"]
    for b in backends:
        key = base64.b64decode(b["key"].encode("utf-8")).decode("utf-8")
        try:
            print("{} [{} @ {}]".format(b["name"], key, b["server"]))
        except:
            pass

    print("\n")
    print("Services")
    print("-----")
    services = config.get("services", [])
    for s in services:
        try:
            print("{} [{}]".format(s["name"], s["path"]))
        except:
            pass


@config.command("backend-add")
@click.argument("name")
@click.argument("server")
@click.argument("key")
@click.pass_context
def _backend_add(ctx, name, server, key):
    """
    Add backend
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    # Encode the key before storing
    key = base64.b64encode(key.encode("utf-8")).decode("utf-8")

    backends = config["backends"]
    missing = True
    for b in backends:
        if ("server" in b) and (b["server"] == server):
            b["name"] = name
            b["server"] = server
            b["key"] = key
            missing = False
            break

    if missing:
        backends.append({"name": name, "server": server, "key": key})

    with open(filename, "w") as fd:
        fd.write(json.dumps(config, indent=4))

    print("Added {}".format(name))


@config.command("backend-remove")
@click.argument("name")
@click.pass_context
def _backend_remove(ctx, name):
    """
    Add backend
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    backends = config["backends"]
    missing = True
    updated = [
        b
        for b in backends
        if ((b.get("server", None) != name) and (b.get("name", None) != name))
    ]

    if len(backends) == len(updated):
        raise Exception("Backend not found: {}".format(name))

    config["backends"] = updated

    with open(filename, "w") as fd:
        fd.write(json.dumps(config, indent=4))

    print("Removed {}".format(name))


@config.command("service-add")
@click.argument("name")
@click.argument("path")
@click.pass_context
def _service_add(ctx, name, path):
    """
    Add URL
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    if "services" not in config:
        config["services"] = []

    services = config["services"]
    missing = True
    for s in services:
        if ("name" in s) and (s["name"] == name):
            s["name"] = name
            s["path"] = path
            missing = False
            break

    if missing:
        services.append({"name": name, "path": path})

    with open(filename, "w") as fd:
        fd.write(json.dumps(config, indent=4))

    print("Added {}".format(name))


@config.command("service-remove")
@click.argument("name")
@click.pass_context
def _service_remove(ctx, name):
    """
    Remove service
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    services = config["services"]
    missing = True
    updated = [s for s in services if s.get("name", None) != name]

    if len(services) == len(updated):
        raise Exception("Service not found: {}".format(name))

    config["services"] = updated

    with open(filename, "w") as fd:
        fd.write(json.dumps(config, indent=4))

    print("Removed {}".format(name))


@click.group()
@click.argument("backend")
@click.pass_context
def show(ctx, backend):
    """
    Access a backend
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    backends = config["backends"]

    relevant = None
    for b in backends:
        if (b.get("server", None) == backend) or (b.get("name", None) == backend):
            relevant = b
            break

    if relevant is None:
        raise Exception("Invalid backend name: {}".format(name))

    # Decode the key before accessing
    key = base64.b64decode(relevant["key"].encode("utf-8")).decode("utf-8")
    relevant["key"] = key

    backend = sdkapi.Backend(relevant)
    ctx.obj["backend"] = backend


@show.command("status")
@click.option("--offset", default=0)
@click.option("--page", default=20)
@click.pass_context
def _status(ctx, offset, page):
    """
    Execution status
    """

    backend = ctx.obj["backend"]
    result = backend.status(offset, page)

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    columns = ["usecase", "nature", "name", "status", "runid", "start_time", "duration"]
    sdkapi.draw(backend, "status", result, columns)


@show.command("usecases")
@click.pass_context
def _usecases(ctx):
    """
    Show available usecases
    """

    backend = ctx.obj["backend"]
    result = backend.usecases()

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    columns = ["customer", "name", "description"]
    sdkapi.draw(backend, "usecases", result, columns)


@show.command("health")
@click.pass_context
def _health(ctx):
    """
    Show system health
    """

    backend = ctx.obj["backend"]
    result = backend.health()

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    sdkapi.draw_health(backend, "health", result)


@show.command("pipelines")
@click.argument("usecase")
@click.pass_context
def _pipelines(ctx, usecase):
    """
    Show available pipelines
    """

    backend = ctx.obj["backend"]
    result = backend.pipelines(usecase)

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    data = result["data"]
    for d in data:
        d["description"] = d["description"][:20]
        runs = sorted(d["runs"], reverse=True)
        if len(runs) > 0:
            d["lastrun"] = runs[0]
            d["runs"] = len(runs)
        else:
            d["lastrun"] = ""
            d["runs"] = 0

    columns = ["usecase", "name", "description", "path", "runs", "lastrun"]
    sdkapi.draw(
        backend,
        "pipelines",
        result,
        columns,
        {
            "Usecase": usecase,
        },
    )


@show.command("run-list")
@click.argument("usecase")
@click.argument("pipeline")
@click.pass_context
def _runs(ctx, usecase, pipeline):
    """
    Show available runs for pipeline
    """

    backend = ctx.obj["backend"]
    result = backend.run_list(usecase, pipeline)

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    result["data"] = sorted(result["data"], key=lambda r: r["start_time"], reverse=True)

    columns = ["runid", "status", "start_time", "end_time"]
    sdkapi.draw(
        backend, "run-list", result, columns, {"Usecase": usecase, "Pipeline": pipeline}
    )


@show.command("run-detail")
@click.argument("usecase")
@click.argument("pipeline")
@click.argument("runid")
@click.pass_context
def _run_detail(ctx, usecase, pipeline, runid):
    """
    Show detail of a given run
    """

    backend = ctx.obj["backend"]
    result = backend.run_detail(usecase, pipeline, runid)

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    sdkapi.draw_run_detail(
        backend,
        "run-detail",
        result,
        {"Usecase": usecase, "Pipeline": pipeline, "RunID": runid},
    )


@show.command("tasks")
@click.argument("usecase")
@click.pass_context
def _tasks(ctx, usecase):
    """
    Show available tasks
    """

    backend = ctx.obj["backend"]
    result = backend.tasks(usecase)

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    data = result["data"]
    for d in data:
        d["description"] = d["description"][:20]
        runs = sorted(d["runs"], reverse=True)
        if len(runs) > 0:
            d["lastrun"] = runs[0]
            d["runs"] = len(runs)
        else:
            d["lastrun"] = ""
            d["runs"] = 0

    columns = ["usecase", "name", "description", "path", "runs", "lastrun"]
    sdkapi.draw(
        backend,
        "tasks",
        result,
        columns,
        {
            "Usecase": usecase,
        },
    )


@show.command("task-run-list")
@click.argument("usecase")
@click.argument("task")
@click.pass_context
def _task_runs(ctx, usecase, task):
    """
    Show available runs for task
    """

    backend = ctx.obj["backend"]
    result = backend.task_run_list(usecase, task)

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    result["data"] = sorted(result["data"], key=lambda r: r["start_time"], reverse=True)

    columns = ["runid", "status", "start_time", "end_time"]
    sdkapi.draw(
        backend, "task-run-list", result, columns, {"Usecase": usecase, "Task": task}
    )


@show.command("task-run-detail")
@click.argument("usecase")
@click.argument("task")
@click.argument("runid")
@click.pass_context
def _task_run_detail(ctx, usecase, task, runid):
    """
    Show detail of a given run
    """

    backend = ctx.obj["backend"]
    result = backend.task_run_detail(usecase, task, runid)

    status = result["status"]
    if status in ["failure", "error"]:
        print("Failure while accessing backend")
        if "error" in result:
            print("Reason:", result["error"])
        return

    sdkapi.draw_run_detail(
        backend,
        "run-detail",
        result,
        {"Usecase": usecase, "Task": task, "RunID": runid},
    )


@click.group()
@click.argument("backend")
@click.argument("service")
@click.pass_context
def featurestore(ctx, backend, service):
    """
    Access featurestore at backend
    """
    try:
        filename = ctx.obj["config"]
        config = json.load(open(filename))
    except:
        traceback.print_exc()
        raise Exception("Invalid config file: {}".format(ctx.obj["config"]))

    backends = config["backends"]

    relevant = None
    for b in backends:
        if (b.get("server", None) == backend) or (b.get("name", None) == backend):
            relevant = b
            break

    if relevant is None:
        raise Exception("Invalid backend name: {}".format(name))

    # Decode the key before accessing
    key = base64.b64decode(relevant["key"].encode("utf-8")).decode("utf-8")
    relevant["key"] = key

    backend = sdkapi.Backend(relevant)
    ctx.obj["backend"] = backend

    # Lookup the service as well..
    services = config["services"]
    relevant = None
    for s in services:
        if s.get("name", None) == service:
            relevant = s
            break

    if relevant is None:
        raise Exception("Invalid backend name: {}".format(name))
    ctx.obj["service"] = s


@featurestore.command("post")
@click.argument("filename")
@click.option("--debug/--no-debug", default=False)
@click.pass_context
def _post(
    ctx,
    filename,
    debug,
):
    """
    Post json to server
    """

    backend = ctx.obj["backend"]
    service = ctx.obj["service"]

    response = fslib.post(
        backend=backend, service=service, filename=filename, debug=debug
    )
    print(response)


@featurestore.command("generate")
@click.option("--debug/--no-debug", default=False)
@click.pass_context
def _generate(ctx, debug):
    """
    Generate sample files
    """

    backend = ctx.obj["backend"]
    service = ctx.obj["service"]

    response = fslib.generate(backend=backend, service=service, debug=debug)
    print(json.dumps(response, indent=4))


@featurestore.command("download")
@click.option("--featuregroup_id", default=None)
@click.option("--run_id", default=None)
@click.option("--debug/--no-debug", default=False)
@click.pass_context
def _spec_download(
    ctx,
    featuregroup_id,
    run_id,
    debug,
):
    """
    Download a featuregroup specification
    """

    if featuregroup_id is None and run_id is None:
        raise Exception("One of the featuregroup_id or run_id should be specified")

    backend = ctx.obj["backend"]
    service = ctx.obj["service"]

    response = fslib.download(
        backend=backend,
        service=service,
        featuregroup_id=featuregroup_id,
        run_id=run_id,
        debug=debug,
    )
    print(json.dumps(response, indent=4))


@featurestore.command(
    "search",
    context_settings=dict(
        ignore_unknown_options=True,
    ),
)
@click.option("--debug/--no-debug", default=False)
@click.argument("params", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def _spec_search(ctx, debug, params):
    """
    Search for a feature spec or run
    """

    backend = ctx.obj["backend"]
    service = ctx.obj["service"]

    def error():
        print("Args should have attr=value format")
        print("Example: ....local spec search name=customer_persona")
        print("You can give django filter query name__icontains=persona")
        return

    for p in params:
        if p.count("=") != 1:
            return error()

    params = {p.split("=")[0]: p.split("=")[1] for p in params}
    if len(params) == 0:
        return error()

    response = fslib.search(
        backend=backend, service=service, params=params, debug=debug
    )

    print(json.dumps(response, indent=4))


api.add_command(config)
api.add_command(show)
api.add_command(featurestore)

###################################
# Add to process
###################################
process.add_command(_version)
process.add_command(_start)
process.add_command(bootstrap_)
process.add_command(envsetup)
process.add_command(api)
process.add_command(test)
process.add_command(sample)
process.add_command(_show_log)


def main():
    process()



"""
Code that goes along with the Airflow tutorial located at:
https://github.com/airbnb/airflow/blob/master/airflow/example_dags/tutorial.py
"""

import os
import sys
from datetime import timedelta, datetime
from airflow import DAG
from airflow.operators.bash_operator import BashOperator
from airflow.operators.python_operator import PythonOperator
from datetime import datetime, timedelta

default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime(2018, 4, 4, 0, 0, 0),
    "email": ["support@scribbledata.io"],
    "email_on_failure": True,
    "email_on_retry": True,
    "retries": 1,
    "retry_delay": timedelta(minutes=5),
    "catchup": False,
    # 'queue': 'bash_queue',
    # 'pool': 'backfill',
    # 'priority_weight': 10,
    # 'end_date': datetime(2016, 1, 1),
}

dag = DAG(
    "contrib-pipeline-v1",
    schedule_interval=timedelta(hours=2),
    catchup=False,
    depends_on_past=False,
    default_args=default_args,
)

cli = os.path.join(os.environ["ENRICH_ROOT"], "manage", "bin", "run_enrichcli.sh")

conf = os.path.join(
    os.environ["ENRICH_CUSTOMERS"], "scribble", "Contrib", "conf", "test.json"
)

t1 = BashOperator(
    task_id="run_test_pipeline",
    bash_command="{} run-pipeline -c {}".format(cli, conf),
    dag=dag,
)


t2 = EmailOperator(
    task_id="notification",
    to="support@scribbledata.io",
    subject="Completed task",
    html_content="Completed ",
    dag=dag,
)

t2.set_upstream(t1)



"""
Redis integration.

"""
import traceback
import json
import pickle
from collections import defaultdict

# Good use of the iterator and del command
# https://stackoverflow.com/questions/21975228/redis-python-how-to-delete-all-keys-according-to-a-specific-pattern-in-python


def bulk_get_data(cache, keys):
    """
    Get the data for keys...
    :param keys: list, list of keys
    :return: list, data at keys
    """

    data = []

    if len(keys) > 0:
        pipe = cache.pipeline()
        for k in keys:
            pipe.hgetall(k)
        data = pipe.execute()

    return data


def bulk_get_keys(cache, pattern, chunk_size=5000):
    """
    Get the keys...
    :param ns: str, namespace i.e your:prefix
    :return: int, cleared keys
    """

    allkeys = []
    cursor = "0"
    while cursor != 0:
        cursor, keys = cache.scan(cursor=cursor, match=pattern, count=chunk_size)
        if keys:
            allkeys.extend(keys)

    return allkeys


def clear(cache, pattern, chunk_size=5000):
    """
    Clears a namespace
    :param ns: str, namespace i.e your:prefix
    :return: int, cleared keys
    """

    count = 0
    cursor = "0"
    while cursor != 0:
        cursor, keys = cache.scan(cursor=cursor, match=pattern, count=chunk_size)
        if keys:
            count += len(keys)
            cache.delete(*keys)

    return count


def clear_atomic(cache, pattern):
    """
    Clears a namespace in redis cache.
    This may be very time consuming.
    :param pattern: str, namespace i.e your:prefix*
    :return: int, num cleared keys
    """
    count = 0
    pipe = cache.pipeline()
    for key in cache.scan_iter(ns):
        pipe.delete(key)
        count += 1
    pipe.execute()

    return count


def write_records_atomic(cache, records, keygenerator, spec, expiry=7 * 86400):
    """
    Serialize and write records

    :param records: list of dictionaries
    :param keygenerator: generates key from dict
    :param spec: what should this function do?
    :return: stats, stats about the redis write

    """

    error = 0
    count = 0
    found = defaultdict(int)
    stats = defaultdict(int)

    exclude = spec.get("exclude", [])  # what columns to exclude
    suffix = spec.get("suffix", "")  # add bits to the key name
    target = spec.get("target", "map")  # what is the target

    # Make sure we know what action to do.
    # map = store a dict
    # map-serialized: serialize the dict and store
    # elements = store individual elements
    # elements-serialized = store individual elements but serialize them..
    #
    if target not in ["map", "map-serialized", "elements", "elements-serialized"]:
        raise Exception(f"Unknown target: {target}")

    if not isinstance(exclude, list):
        raise Exception(f"Unknown type of exclude {type(exclude)}")

    pipe = cache.pipeline()
    for r in records:

        try:

            # First generate the key using the entire record
            if isinstance(keygenerator, str):
                key = keygenerator % r
            elif callable(keygenerator):
                key = keygenerator(r)
            else:
                raise Exception("Unknown key generator")

            # What are the available columns?
            columns = spec.get("columns", list(r.keys()))

            # What should I exclude?
            columns = [c for c in columns if c not in exclude]

            # Get what I should write
            towrite = {k: v for k, v in r.items() if k in columns}

            if len(towrite) == 0:
                stats["error-empty"] += 1
                continue

            # Handle the simple cases...
            if spec["target"] in ["map", "map-serialized"]:
                key += suffix
                if key in found:
                    stats["duplicates"] += 1
                    continue
                if target == "map-serialized":  # serialize
                    value = pickle.dumps(towrite)
                    pipe.set(key + "_pickled", value)
                else:
                    pipe.hmset(key, towrite)
                found[key] = 1

                # In all cases set the expiry...
                if expiry > 0:
                    pipe.expire(key, expiry)

                stats["valid"] += 1
                continue

            # Handle the list case...
            if spec["target"] in ["elements", "elements-serialized"]:
                for col in list(towrite.keys()):
                    colkey = f"{key}:{col}{suffix}"
                    # Have we seen this key?
                    if colkey in found:
                        stats["duplicates"] += 1
                        continue

                    value = towrite[col]

                    # We now have to set the list...
                    if target == "elements-serialized":
                        value = pickle.dumps(value)
                        pipe.set(colkey + "_pickled", value)
                    elif isinstance(value, list):
                        pipe.delete(colkey)  # clear all elements
                        pipe.rpush(colkey, *value)
                    elif isinstance(value, dict):
                        pipe.hmset(colkey, value)
                    else:
                        raise Exception(
                            f"Unknown element type {type(value)} for {colkey}"
                        )

                    found[colkey] += 1
                    stats["valid"] += 1

        except Exception as e:
            stats["error"] += 1
            stats["message"] = str(e)
            break

        count += 1

    pipe.execute()

    return dict(stats)


def write_df_dict(cache, df, keygenerator, spec, expiry=7 * 86400):
    """
    Serialize and write dataframe

    :param df: dataframe to write
    :param keygenerator: generates key from dict
    :param spec: how should the dict be handled?
    :param expiry: expiry in seconds (default 7 days)
    :return: stats, stats about the redis write

    Write selected columns
    spec = {
        "target": "map",
        "columns": ["key1", "key2"]
    ]

    Write selected columns but serialize them if they
    are not of the standard datatype
    spec = {
        "target": "map-serialized",
        "columns": ["key1", "key2"]
    ]


    Write each of the columns to a separate key
    spec = {
        "target": "elements",
        "columns": ["key1", "key2"]
    ]


    """

    # Generate records
    records = df.to_dict("records")

    return write_records_atomic(cache, records, keygenerator, spec, expiry)



"""
This developer kit is meant for advanced users of the `Scribble Enrich Feature Store`. 
This includes command line scripts to create and manage Enrich modules and APIs to 
write new modules on the platform::
  
      ______           _      __       _____ ____  __ __
     / ____/___  _____(_)____/ /_     / ___// __ \\/ //_/
    / __/ / __ \\/ ___/ / ___/ __ \\    \\__ \\/ / / / ,<   
   / /___/ / / / /  / / /__/ / / /   ___/ / /_/ / /| |  
  /_____/_/ /_/_/  /_/\\___/_/ /_/   /____/_____/_/ |_|  
  
"""
import os

VERSION = "3.7.6"


def _get_git_revision(path):
    revision_file = os.path.join(path, "refs", "heads", "master")
    if os.path.exists(revision_file):
        with open(revision_file) as fh:
            return fh.read().strip()[:7]


def get_revision():
    """
    :returns: Revision number of this branch/checkout, if available. None if
        no revision number can be determined.
    """
    package_dir = os.path.dirname(__file__)
    checkout_dir = os.path.normpath(os.path.join(package_dir, os.pardir))
    path = os.path.join(checkout_dir, ".git")
    if os.path.exists(path):
        return _get_git_revision(path)


def get_version():
    base = VERSION
    if __build__:
        base = "%s (%s)" % (base, __build__)
    return base


__build__ = get_revision()
__version__ = VERSION

from .core import *
from .package import *
from .lib import *
from .commands import *
from .tasks import *
from .services import *
from .datasets import *
from .quality import *
from . import featurestore, realtime, contrib, utils, notebook, feature_compute



import os
import sys
import json
import logging
from datetime import datetime
import platform
import distro
import getpass
import logging
from collections import OrderedDict
import logging

from . import log

logger = logging.getLogger("app")


class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        else:
            return super().default(obj)


#######################################################
# Config object to handle a number of helper functions
#######################################################
class Config(object):
    """
    Context object having a number of capabilities including
    log management, files and path management, credentials
    lookup etc.

    """

    def __init__(self, params, *args, **kwargs):
        """
        Initialize the config (context) object.

        Args:
          params (dict): Only two keys are supported. debug (bool) prints
                         extra information. It is useful for debugging loading.
                         settings is the path to the settings file.
        """
        super().__init__(*args, **kwargs)
        self.debug = params.get("debug", False)
        self.settings = params.get("settings", None)
        self.readonly = params.get("readonly", False)
        self.show_metadata = params.get("show_metadata", True)

        if self.settings is not None:
            self.settings = os.path.expanduser(self.settings)

        self.now = datetime.now()
        self.conf = {"credentials": {}, "libraries": []}
        self.stats = {
            "platform": {
                "node": platform.node(),
                "os": platform.system(),
                "release": platform.release(),
                "processor": platform.processor(),
                "python": platform.python_version(),
                "distribution": distro.linux_distribution(),
            },
            "user": getpass.getuser(),
        }

        self.state = self.get_default_metadata()
        self.state.update({"outputs": [], "notes": [], "message": ""})

    def get_default_metadata(self):
        """
        Get reusable metadata dict with some standard fields.
        """
        metadata = {
            "schema": "standalone:command",
            "version": "1.0",
            "timestamp": datetime.now().replace(microsecond=0).isoformat(),
            "command": {
                "name": "Unknown",
                "status": "initial",
                "description": "",
                "host": self.stats["platform"]["node"],
                "stats": self.stats,
                "pid": os.getpid(),
                "start_time": self.now,
                "end_time": self.now,
                "cmdline": list(sys.argv),
            },
        }

        return metadata

    def configure(self, settings=None):
        """
        Configure the config object

        Args:
           settings (str): Path to the settings file

        """
        if settings is None:
            settings = self.settings

        if (settings is None) or (not os.path.exists(settings)):
            # Quietly return. Do nothing..
            raise Exception("Settings path invalid or undefined: {}".format(settings))

        try:
            loaded_conf = json.load(open(settings))
        except:
            logger.exception("Invalid settings file")
            loaded_conf = None

        if loaded_conf is None:
            return

        # Materialize the paths...
        self.enrich_root = loaded_conf.get("enrich_root")
        self.enrich_root = os.path.expanduser(self.enrich_root)
        self.enrich_root = os.path.abspath(self.enrich_root)

        libraries = loaded_conf.get("libraries", [])
        libraries = [self.get_file(l) for l in libraries]
        self.conf["libraries"] = libraries

        # Insert the rest
        defaults = {
            "enrich_data": "%(enrich_root)s/data",
            "enrich_customers": "%(enrich_root)s/customers",
            "runid": "run-%Y%m%d-%H%M%S-%(user)s",
            "command_root": "%(enrich_data)s/_commands",
            "rundir": "%(command_root)s/%(hostname)s/%(runid)s",
            "outputdir": "%(rundir)s/output",
            "log": "%(rundir)s/log.json",
            "metadata": "%(rundir)s/metadata.json",
            "args": {},
        }

        for name in [
            "enrich_data",
            "enrich_customers",
            "runid",
            "command_root",
            "rundir",
            "outputdir",
            "log",
            "metadata",
        ]:
            default = defaults[name]
            try:
                final = loaded_conf.get(name, default)
                final = self.get_file(final)
            except:
                error = "Could not resolve: {} ({})".format(name, final)
                raise Exception(error)

            setattr(self, name, final)
            self.state["command"][name] = final

        for name in ["args"]:
            default = defaults[name]
            try:
                final = loaded_conf.get(name, default)
            except:
                error = "Could not resolve: {} ({})".format(name, final)
                raise Exception(error)

            setattr(self, name, final)

        # Set environment variables.....
        os.environ["ENRICH_ROOT"] = self.enrich_root
        os.environ["ENRICH_CUSTOMERS"] = self.enrich_customers
        os.environ["ENRICH_DATA"] = self.enrich_data

    def get_file(self, path):
        """
        Resolve the abstract specification into a full path

        Args:
           path (str): Path specification

        Example::

           %(enrich_data)/acme/Projects/commands

        This is resolved into `/home/ubuntu/enrich/data/acme/Projects/commands`

        """
        path = path.replace("%(", "%%(")  # This will make the strftime handling safe
        path = self.now.strftime(path)
        path = path % {
            "enrich_root": self.enrich_root,
            "enrich_data": getattr(self, "enrich_data", ""),
            "enrich_customers": getattr(self, "enrich_customers", ""),
            "user": self.stats["user"],
            "username": self.stats["user"],
            "hostname": self.stats["platform"]["node"],
            "node": self.stats["platform"]["node"],
            "runid": getattr(self, "runid", ""),
            "rundir": getattr(self, "rundir", ""),
            "outputdir": getattr(self, "outputdir", ""),
        }
        return path

    def get_relative_path(self, path, what="enrich_root"):

        if what == "enrich_root":
            return os.path.relpath(path, self.enrich_root)
        elif what == "enrich_data":
            return os.path.relpath(path, self.enrich_data)
        else:
            raise Exception("relative path should specify the root")

    def add_output(self, params):
        """
        Add file to the command output

        Args:
           params (dict): file parameters
              File parameters includes a name of the output, relative path
              to the file, and a description for the output

        """
        required = ["name", "description", "path"]
        missing = [n for n in required if n not in params]
        if len(missing) > 0:
            raise Exception("Missing parameters for the output: {}".format(missing))

        if os.path.isabs(params["path"]):
            raise Exception(
                "Path specified for output {} should be relative".format(params["name"])
            )

        for o in self.state["outputs"]:
            if o["name"] == params["name"]:
                raise Exception(
                    "An output already exists with the name: {}".format(params["name"])
                )

        params["fullpath"] = self.get_file("%(outputdir)s/{}".format(params["path"]))
        params["relpath"] = self.get_relative_path(params["fullpath"])

        # Now append
        self.state["outputs"].append(params)

    def get_output_handle(self, name, *args):
        """
        Get a file descriptor to the named output.

        Args:
           name (str): Name of the output.
              This output was previously added using `add_output` call.

        """
        for o in self.state["outputs"]:
            if o["name"] == name:
                try:
                    os.makedirs(os.path.dirname(o["fullpath"]))
                except:
                    pass
                return open(o["fullpath"], *args)

        raise Exception("Could not find output {}".format(name))

    def get_library_paths(self):
        return self.conf["libraries"]

    def get_args(self, cmd):
        """
        Get default arguments to the command specified in the settings file

        Args:
           cmd (str): Name of the command

        """
        logger = logging.getLogger("app")
        if not hasattr(self, "args"):
            if self.debug:
                logger.error("args not present")
            return {}

        if not isinstance(self.args, dict):
            if self.debug:
                logger.error("args not a dictionary")
            return {}

        return self.args.get(cmd, {})

    # Update the state
    def set_status(self, value):
        self.state["command"]["status"] = value

    def set_message(self, value):
        self.state["message"] = value

    def dump_state(self):

        # Update file statistics
        for o in self.state["outputs"]:
            path = o["fullpath"]
            st = os.stat(path)
            o["size"] = st.st_size

        metadata = OrderedDict(
            [
                ("schema", self.state["schema"]),
                ("version", self.state["version"]),
                ("timestamp", self.state["timestamp"]),
                ("command", self.state["command"]),
                ("message", self.state["message"]),
                ("outputs", self.state["outputs"]),
                ("notes", self.state["notes"]),
                ("log", os.path.relpath(self.log, self.rundir)),
            ]
        )

        if not self.readonly:
            with open(self.metadata, "w") as fd:
                json.dump(metadata, fd, indent=4, cls=CustomEncoder)
        elif self.show_metadata:
            print(json.dumps(metadata, indent=4, cls=CustomEncoder), file=sys.stderr)

        if self.debug:
            logger.debug("Dumped execution metadata")

    # Logging start and end
    def start_run(self, cmd):

        # Override with parameters from settings
        if not self.readonly:
            args = self.get_args(cmd)
            self.readonly = args.get("readonly", self.readonly)
            self.show_metadata = args.get("show_metadata", self.show_metadata)

        # Setup logging
        if self.readonly:
            log.logging_start()
        else:
            log.logging_start(self.log)

        logger = logging.getLogger("app")
        if self.debug:
            logger.debug("Starting run")

        # Dump the state at the beginning
        self.dump_state()

    def end_run(self):

        self.state["command"]["end_time"] = datetime.now()
        self.dump_state()

        # Dump status
        logger = logging.getLogger("app")

        if self.debug:
            logger.debug("Ending run")
        log.logging_end()



"""
"""
import os
import sys
import json
import copy

from ..lib import read_siteconf


class Service(object):
    """
    Base class for all services.

    Each class of services extends this base class to achieve its
    performance and other objectives. For example, `data quality
    service` (DataSourceService) that is provided by the `Contrib`
    application from `Scribble` customer.

    """

    def __init__(self, conf, *args, **kwargs):
        """
        Initialize the service manager
        """
        if "name" not in conf:
            raise Exception("Invalid configuration: Every service requires a name")

        self.conf = conf
        """
        Configuration of this service. 
        
        This is interpreted by the subclass functions 
        """

        self.name = self.conf["name"]
        """
        Name of the service. 
        """

        self.context = {
            "enrich_root": os.environ["ENRICH_ROOT"],
            "enrich_root_dir": os.environ["ENRICH_ROOT"],
            "enrich_data": os.environ["ENRICH_DATA"],
            "enrich_data_dir": os.environ["ENRICH_DATA"],
        }
        """
        Context for path resolution
        """

    def get_file(self, path, extra={}):
        """
        Resolve a path
        """

        fullcontext = copy.copy(self.context)
        fullcontext.update(extra)
        return path % fullcontext

    def get_credentials(self, name):
        """
        Look up the credentials file
        """

        siteconf = read_siteconf()
        if siteconf is None:
            raise Exception("Could not find siteconf")

        if ("credentials" not in siteconf) or (name not in siteconf["credentials"]):
            raise Exception("missing credentials")

        return siteconf["credentials"][name]



import os
from jinja2 import Environment, FileSystemLoader, PackageLoader, meta

thisdir = os.path.dirname(__file__)
default_templatedir = os.path.abspath(os.path.join(thisdir, "..", "templates"))

# from prompt_toolkit.token import Token
# from prompt_toolkit.styles import style_from_dict
# style = style_from_dict({
#    Token.Toolbar: '#ffffff bg:#333333',
# })


##########################
# Jinja2 helper functions...
##########################
def get_template_path(filename, templatedir=None):

    if os.path.isabs(filename):
        return filename

    if templatedir is None:
        templatedir = default_templatedir

    # Check if relative path is specified, then use it.
    # if os.path.exists(filename):
    #   return filename

    return os.path.join(templatedir, filename)


def list_templates(suffix=".node.template"):

    mapping = {}
    for t in os.listdir(default_templatedir):
        if t.endswith(suffix):
            path = os.path.join(default_templatedir, t)
            mapping[t.replace(suffix, "")] = path
    return mapping


def render(tpl_path, context):
    path, filename = os.path.split(tpl_path)
    return (
        Environment(loader=FileSystemLoader(path or "./"))
        .get_template(filename)
        .render(context)
    )


def get_variables(tpl_path):
    path, filename = os.path.split(tpl_path)
    env = Environment(loader=FileSystemLoader(path or "./"))
    template_source = env.loader.get_source(env, filename)[0]
    parsed_content = env.parse(template_source)
    return meta.find_undeclared_variables(parsed_content)


def write_rendered_file(templatefile, target, filename, params):

    if not templatefile.startswith("/"):
        templatefile = get_template_path(templatefile)

    if not os.path.exists(templatefile):
        raise Exception("Template file doesnt exist")

    content = render(templatefile, params)
    try:
        os.makedirs(target)
    except:
        pass

    with open(os.path.join(target, filename), "w") as fd:
        fd.write(content)
    return content



import json
from sqllineage.runner import LineageRunner

from .sample import *


def get_lineage_of_query(engine, sql):
    """
    Given an engine and an SQL expression, find the input
    and output tables. Replace <default> with database
    name from the engine
    """
    dependencies = []
    result = LineageRunner(sql, verbose=True)

    database = engine.url.database

    # Get input tables..
    if len(result.source_tables) > 0:
        tables = [str(t).replace("<default>", database) for t in result.source_tables]

        tables = {
            t.split(".")[-1]: {"text": t.replace(".", " "), "table": t} for t in tables
        }
        dependencies.append(
            {
                "type": "db",
                "nature": "input",
                "objects": tables,
            }
        )

    # Get output dependencies
    if len(result.target_tables) > 0:
        tables = [str(t).replace("<default>", database) for t in result.target_tables]
        tables = {
            t.split(".")[-1]: {"text": t.replace(".", " "), "table": t} for t in tables
        }

        dependencies.append({"type": "db", "nature": "output", "objects": tables})

    return dependencies

class SafeEncoder(json.JSONEncoder):

    def default(self, obj):
        try:
            if isinstance(obj, (datetime, date)):
                result = obj.isoformat()
            else:
                result = super().default(obj)
        except:
            result = str(obj)

        return result




import os
import json
import traceback

from flask import Flask, render_template, request
from flask_multistatic import MultiStaticFlask

app = MultiStaticFlask(__name__)


@app.route("/")
def main():

    print(request.args)
    # First get the arguments if any
    params = {}
    for name in request.args:
        valuelist = request.args.getlist(name)
        params[name] = valuelist

    try:
        skin = app.config["skin"]
        data = skin.render(params)
    except:
        traceback.print_exc()
        data = [
            {
                "target": "result",
                "rendered": "<p class='lead'>Error while rendering</p>",
            }
        ]

    return render_template("index.html", data=data)


def shutdown_server():
    func = request.environ.get("werkzeug.server.shutdown")
    if func is None:
        raise RuntimeError("Not running with the Werkzeug Server")
    func()


@app.route("/shutdown")
def shutdown():
    shutdown_server()
    return "Server shutting down..."


# Custom static data
@app.route("/app/static/<path:filename>")
def custom_static(filename):
    for p in app.config["static_paths"]:
        fullpath = os.path.join(p, filename)
        if os.path.exists(fullpath):
            return send_from_directory(p, filename)


def create_app(conf):

    # => Set the static folder...
    static_paths = conf.pop("static_paths")
    static_paths.insert(0, os.path.join(app.root_path, "static"))
    app.static_folder = static_paths

    app.config.update(conf)
    return app


if __name__ == "__main__":
    app.run(debug=True)



import copy
import json
import logging

logger = logging.getLogger("app")

__all__ = ["Task"]


class Task(object):
    """
    This is the task base class. This is 'executed' by
    a 'runner'
    """

    NAME = "BaseTask"

    def __init__(self, *args, **kwargs):
        self.config = kwargs.pop("config")
        self.name = "BaseTask"
        self.description = "Baseclass of tasks"
        self.version = "1.0"
        self.testdata = {"data": {}}

        self.supported_extra_args = []
        """Extra arguments that can be passed on the commandline to 
        the module. 

        self.extra_args = [
               {
                  "name": "jobid",
                  "description": "Export JobID",
                  "type": "str",
                  "required": True,
                  "default": "22832",
               }
        ]
        """

    #################################################
    # Configuration the Module
    #################################################
    def preload_clean_args(self, args):
        """
        Clean the arguments before using them

        Parameters
        ----------
        args: Args specified in the config file
        """

        if not self.config.enable_extra_args:
            return args

        # Override the defaults with the parameters
        # Three sources for a given args attribute in the order of
        # priority
        #
        # 1. command line
        # 2. Configuration
        # 3. Default

        readonly = self.config.readonly
        cmdargs = self.config.get_cmdline_args()
        supported_extra_args = self.supported_extra_args

        # Collect the final values for all the defaults
        final = {}
        for a in supported_extra_args:
            name = a["name"]
            cmdlabel = "{}:{}".format(self.name, name)
            default = a["default"]
            required = a["required"]
            if required and not readonly:
                # Required and specified on the command line...
                if cmdlabel not in cmdargs:
                    logger.error(
                        "Invalid configuration",
                        extra={
                            "transform": self.name,
                            "data": "Variable {} for task {} is required but missing in command line".format(
                                name, self.name
                            ),
                        },
                    )
                    raise Exception("Invalid configuration")
                final[name] = cmdargs[cmdlabel]
            elif cmdlabel in cmdargs:
                # Optional but specified on the command line
                final[name] = cmdargs[cmdlabel]
            elif name in args:
                # Priority 2
                final[name] = args[name]
            else:
                # Priority 3
                final[name] = default

        # => Log what has happened...
        msg = ""
        for k, v in final.items():
            msg += "{}: {}\n".format(k, v)
        logger.debug(
            "Configuration has been overridden",
            extra={"transform": self.name, "data": msg},
        )

        args.update(final)

        return args

    def preload_validate_self(self):
        """
        Check whether definition of the task is fine

        """
        supported_extra_args = self.supported_extra_args
        if not isinstance(supported_extra_args, list):
            logger.error(
                "Invalid configuration. Expected {}, actual {}".format(
                    "list", str(type(supported_extra_args))
                ),
                EXTRA=self.config.get_extra(
                    {
                        "transform": self.name,
                    }
                ),
            )
            raise Exception("Supported extra argument should be a list of dicts")

        for i, a in enumerate(supported_extra_args):
            if not isinstance(a, dict):
                logger.error(
                    "Invalid configuration. Entry {} of supported_extra_args: Expected {}, actual {}".format(
                        i, "dict", str(type(a))
                    ),
                    EXTRA=self.config.get_extra(
                        {
                            "transform": self.name,
                        }
                    ),
                )
                raise Exception("Supported extra argument should be a list of dicts")

            required = ["name", "description", "default", "required"]
            missing = [x for x in required if x not in a]
            if len(missing) > 0:
                logger.error(
                    "Invalid configuration. Entry {} of supported_extra_args: Missing attributes: {}".format(
                        i, ", ".join(missing)
                    ),
                    extra=self.config.get_extra(
                        {"transform": self.name, "data": str(a)}
                    ),
                )
                raise Exception("Missing attributes in one of the supported extra args")

    def get_default_metadata(self, state):
        """
        Get reuse metadata dict with some standard fields.
        """
        metadata = {
            "schema": "standalone:task",
            "version": "1.0",
            "timestamp": datetime.now().replace(microsecond=0).isoformat(),
            "task": {
                "usecase": self.config.usecase["org"]["name"],
                # Backward compatability
                "customer": self.config.usecase["org"]["name"],
                "name": state.name,
                "description": self.config.description,
                "host": state.stats["platform"]["name"],
                "runid": state.runid,
                "pid": state.pid,
                "start_time": state.start_time.replace(microsecond=0).isoformat(),
                "end_time": state.end_time.replace(microsecond=0).isoformat(),
                "versionmap": self.config.get_versionmap(),
                "stats": state.stats,
                "cmdline": list(sys.argv),
            },
        }

        return metadata

    def preload_validate_conf(self, conf):
        """
        Check whether this configuration is even valid?

        Parameters
        ----------
        conf: Module configuration specified in the config file
        """

        if not (isinstance(conf, dict) and len(conf) > 0):
            logger.error(
                "Conf is not a valid dictionary",
                extra=self.config.get_extra({"transform": self.name}),
            )
            raise Exception("Conf is not a valid dictionary")

        if self.version != conf.get("version", "1.0"):
            logger.error(
                "Version mismatch",
                extra=self.config.get_extra({"transform": self.name}),
            )
            raise Exception("Version mismatch between config and module")

    def get_supported_extra_args(self):
        """
        What extra command line arguments are supported

        Returns
        -------
        list: List of arg specifications

        see supported_extra_args
        """
        return copy.copy(self.supported_extra_args)

    def configure(self, spec):
        """
        Configure this task
        """

        # => Check whether internal configuration is valid
        self.preload_validate_self()

        # => Check whether passed configuration is fine
        self.preload_validate_conf(spec)

        # Make a copy of the spec
        self.spec = copy.deepcopy(spec)

        # Override attributes from the spec
        for attr in ["name", "output_version", "test", "enable"]:
            if (attr in spec) and hasattr(self, attr):
                setattr(self, attr, spec[attr])

        # Clean and load args
        args = spec.get("args", {})
        self.args = self.preload_clean_args(args)

    # => Validation code
    def validate_conf(self, what, state):
        pass

    def validate_args(self, what, state):
        pass

    def validate_results(self, what, state):
        pass

    def validate_testdata(self, what, state):

        if not hasattr(self, "testdata"):
            raise Exception("testdata element is missing")

        testdata = self.testdata
        if (
            not isinstance(testdata, dict)
            or ("conf" not in testdata)
            or ("data" not in testdata)
        ):
            raise Exception("testdata should be dictionary with conf and data elements")

    def validate(self, what, state):
        """
        Validate various aspects of the task state,
        configuration, and data. Dont override this function. Override
        specific functions such as `validate_args`.

        Args:
            what (str): What should be validated? args, conf, results etc.

        Returns:

             Nothing

        Raises:
             Exception ("Validation error")

        """

        func = getattr(self, "validate_" + what, None)
        if hasattr(func, "__call__"):
            return func(what, state)
        else:
            logger.error(
                "Cannot find function to validate {}".format(what),
                extra=self.config.get_extra({"transform": self.name}),
            )
            raise Exception("Validation error")

    def initialize(self):
        """
        Initialize the task by connecting to database etc.
        """
        pass

    def get_credentials(self, name):
        """
        Helper function to access siteconf for credentials
        """
        if not hasattr(self.config, "siteconf"):
            raise Exception("No siteconf")

        siteconf = self.config.siteconf
        if not isinstance(siteconf, dict):
            raise Exception("Invalid siteconf format")

        try:
            credentials = siteconf["credentials"][name]
        except:
            logger.exception(
                "Cannot acquire credentials",
                extra=self.config.get_extra({"transform": self.name}),
            )
            raise Exception("Missing/invalid credentials")

        return credentials

    def run(self, *args, **kwargs):
        """
        Main execution func. Override this
        """
        raise Exception("Should be implemented by the derived class")


if __name__ == "__main__":
    print(vars())
    subclasses = vars()["Task"].__subclasses__()
    for cls in subclasses:
        print(cls.NAME)



"""
Workflow
--------

This module has simple modules for building a realtime transform
pipeline. They are called workflows here to avoid confusion with
'pipeline' used elsewhere.

"""

import traceback
import logging

logger = logging.getLogger("app")

from .transforms import TransformBase


class State(object):
    """
    Internal state objects. Transform classes use this
    state object to store and retrieve state during
    execution.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.frames = {}

    def put(self, name, value):
        self.frames[name] = value

    def get(self, name, default=None):
        return self.frames.get(name, default)

    def dumps(self):
        return str(self.frames.keys())


class WorkflowBase(object):
    """
    Base class to define a workflow. This class registers new
    transforms. Right now it is a simple linear flow. We expect to
    support DAGs in future.

    """

    def __init__(self, *args, **kwargs):
        """ """
        self.config = kwargs.get("config", {})
        self.debug = kwargs.get("debug", False)
        self.transforms = []
        self.logger = logging.getLogger()

    def add(self, specs, **kwargs):

        if not isinstance(specs, list):
            raise Exception("Add expects a list of specifications")

        globalargs = self.config.get("args", {})

        for index, spec in enumerate(specs):

            if issubclass(spec, TransformBase):
                transform = spec
                name = spec.__name__  # class
            elif isinstance(spec, dict):
                if "transform" not in spec:
                    raise Exception(
                        "Spec dictionary should include a 'transform' class"
                    )
                transform = spec["transform"]
                name = spec.get("name", transform.name)
            else:
                raise Exception(f"Unexpected data type {type(spec)}")

            # Instantiate the transform...
            t = transform(**kwargs)

            # Configure it with the new args...
            args = globalargs.get(name, {})
            args["name"] = name
            t.configure(pipeline=self, args=args)

            # Do any other initialization such as connecting to the
            # database..
            t.initialize()

            # Add it and keep it ready for use..
            self.transforms.append(t)

    def process(self, streamrow={}):
        """
        Process one input record obtained from the stream such as
        Kafka.

        """

        state = State()
        state.put("streamrow", streamrow)

        for t in self.transforms:
            try:
                if not t.enable:
                    continue
                t.process(state)
            except:
                logger.exception("Unable to process")
                raise

        return state

    def get_extra(self):
        """
        Generate pipeline metadata
        """
        extra = {}
        if self.debug:
            versions = [str(t) for t in self.transforms]
            extra = {"versions": versions}
        return extra



import os
import sys
import smtplib
import mimetypes
from email import encoders
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage


def attach_files(msg, attachments=[]):

    # https://docs.python.org/3.4/library/email-examples.html
    for path in attachments:

        if (not isinstance(path, str)) or (not os.path.isfile(path)):
            continue

        ctype, encoding = mimetypes.guess_type(path)
        if ctype is None or encoding is not None:
            ctype = "application/octet-stream"
        maintype, subtype = ctype.split("/", 1)
        if maintype == "text":
            with open(path) as fp:
                attach = MIMEText(fp.read(), _subtype=subtype)
        elif maintype == "image":
            with open(path, "rb") as fp:
                attach = MIMEImage(fp.read(), _subtype=subtype)
        elif maintype == "audio":
            with open(path, "rb") as fp:
                attach = MIMEAudio(fp.read(), _subtype=subtype)
        else:
            with open(path, "rb") as fp:
                attach = MIMEBase(maintype, subtype)
                attach.set_payload(fp.read())
            encoders.encode_base64(attach)

        filename = os.path.basename(path)
        attach.add_header("Content-Disposition", "attachment", filename=filename)
        msg.attach(attach)


def send_html_email(content, sender, receivers, subject, attachments=[]):
    """
    Send HTML email to target audience
    """
    msg = MIMEMultipart()
    msg.attach(MIMEText(content, "html"))

    msg["From"] = sender
    msg["To"] = ", ".join(receivers)
    msg["Subject"] = subject

    # Add payload
    if isinstance(attachments, list) and len(attachments) > 0:
        attach_files(msg, attachments)

    server = os.environ["EMAIL_HOST"]
    username = os.environ["EMAIL_HOST_USER"]
    password = os.environ["EMAIL_HOST_PASSWORD"]
    port = os.environ.get("EMAIL_PORT", "587")
    timeout = os.environ.get("EMAIL_TIMEOUT", "30")

    # Collect the body
    text = msg.as_string()

    smtp = smtplib.SMTP(server, port=int(port), timeout=int(timeout))
    smtp.starttls()
    smtp.login(username, password)
    smtp.sendmail(sender, receivers, text)



import os
import sys
import copy
import json
import re
from enrichsdk import Sink
import logging

logger = logging.getLogger("app")


class TableSink(Sink):
    """
    Transform to dump dataframes in state into files.

    Parameters specific to this module include:

        * sink: A dictionary of dataframe names and how to output them. It has a number of attributes:

            * type: Output type. Only 'table' value is supported for this
              option right now.
            * filename: Output filename. You can use default parameters such
              runid

          The name of the dataframe can be a regular expression allowing you
          specify a simple rule for arbitrary number of frames.

    Example::

        ....
        "transforms": {
            "enabled": [
                ...
                {
                    "transform": "TableSink",
                    "args": {
                        "sink": {
                            "article": {
                                "frametype": "pandas",
                                "filename": "%(output)s/%(runid)s/article.csv",
                                "params": {
                                    "sep": "|"
                                }
                            },
                            ...
                        }
                    }
                    ...
                }
            ]
         }
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "TableSink"

        self.testdata = {
            "conf": {
                "args": {
                    "sink": {
                        "cars": {
                            "frametype": "pandas",
                            "filename": "%(output)s/cars_revised.csv",
                            "params": {"sep": ","},
                        }
                    }
                }
            },
            "data": {
                "carmodel": {
                    "transform": "CarModel",
                    "filename": "cars.csv",
                    "params": {"sep": ","},
                    "state": {
                        "params": [
                            {
                                "type": "args",
                                "transform": "TableSink",
                                "args": {"save": False, "rows": 124},
                            }
                        ]
                    },
                }
            },
        }

    def preload_clean_args(self, args):
        """
        Check to make sure that the arguments
        is consistent with the specification mentioned
        above
        """

        # Sanity check...
        assert isinstance(args, dict)
        assert len(args) > 0

        # Backward compatability
        if "sink" not in args:
            args = {"sink": args}

        args = super().preload_clean_args(args)

        assert "sink" in args
        sink = args["sink"]
        assert isinstance(sink, dict)
        assert len(sink) > 0

        for pattern, detail in sink.items():

            if (
                ("type" in detail)
                and ("frametype" not in detail)
                and (detail["type"] == "table")
            ):
                detail["frametype"] = "pandas"

            if ("frametype" not in detail) or (detail["frametype"] != "pandas"):
                logger.error(
                    "Invalid configuration. Only pandas table source supported by this sink transform",
                    extra=self.config.get_extra({"transform": self.name}),
                )
                raise Exception("Invalid configuration")

            if (
                ("filename" not in detail)
                or (not isinstance(detail["filename"], str))
                or ("params" not in detail)
                or (not isinstance(detail["params"], dict))
            ):
                logger.error(
                    "Invalid args. Filename (string) and params (dict) are required",
                    extra=self.config.get_extra({"transform": self.name}),
                )
                raise Exception("Invalid configuration")

            detail["root"] = self.config.enrich_data_dir

            tags = detail.get("tags", [])
            if isinstance(tags, str):
                tags = [tags]
            detail["tags"] = tags

            sortcols = detail.get("sort", [])
            if isinstance(sortcols, str):
                sortcols = [sortcols]
            detail["sort"] = sortcols

        return args

    def validate_args(self, what, state):
        """
        Extra validation of the arguments
        """
        sink = self.args["sink"]

        assert isinstance(sink, dict)
        for pattern, detail in sink.items():
            assert ("frametype" in detail) and (detail["frametype"] == "pandas")
            assert "filename" in detail
            assert "params" in detail

    def process(self, state):
        """
        Execute the tablesink specification
        """

        # => What exists in the state?
        available_frames = state.get_frame_list()

        # => First construct input for the pandasframe
        extra = {}
        write_input = {}
        args_input = {}
        framecls = self.config.get_dataframe("pandas")
        skipped = []

        sink = self.args["sink"]
        for pattern in sink:
            # The pattern could be precise dataframe name or could be
            # regular expression.
            regex = re.compile("^{}$".format(pattern))
            frames = [
                m.group(0) for f in available_frames for m in [regex.search(f)] if m
            ]

            for f in frames:
                # For each dataframe that is in the system

                detail = state.get_frame(f)

                # => Are there any extra instructions?
                overrides = self.frame_get_overrides(detail)

                # => Materialize the path...
                filename = sink[pattern]["filename"]
                filename = self.config.get_file(
                    filename, create_dir=True, extra={"frame_name": f}
                )
                # Collect all column information
                extra[f] = {
                    "columns": self.collapse_columns(detail),
                    "notes": self.collapse_notes(detail),
                    "descriptions": self.collapse_descriptions(detail),
                    "overrides": overrides,
                }

                # Which dataframe
                df = detail["df"]

                # Get the
                frametype = detail["frametype"]

                # Order the dataframe if it is needed
                sortcols = sink[pattern]["sort"]
                if len(sortcols) > 0:
                    df = framecls.sort_values(df, sortcols, ascending=False)
                params = sink[pattern].get("params", {})

                # Should I be writing this csv?
                save = params.get("save", True)
                save = overrides.get("save", save)

                if not save:
                    skipped.append(f)

                write_input[f] = {
                    "save": save,
                    "frametype": frametype,
                    "pattern": pattern,
                    "df": df,
                    "filename": filename,
                    "params": params,
                }

                args_input[f] = copy.copy(sink[pattern])
                args_input[f]["filename"] = filename

        if len(skipped) > 0:
            logger.warning(
                "Not saving {} tables".format(len(skipped)),
                extra={"transform": self.name, "data": skipped},
            )

        # => Write output details
        framecls.write(args_input, write_input)

        for name in write_input:

            detail = write_input[name]

            # => Insert columns and tags
            pattern = detail["pattern"]

            #
            detail["params"]["tags"] = sink[pattern]["tags"]

            # Incorporate columns, notes and description
            additional_params = extra[name]
            overrides = additional_params.pop("overrides", {})

            detail["params"].update(additional_params)

            # Insert any overrides provided in the state
            if "rows" in overrides:
                detail["params"]["components"][0]["rows"] = overrides["rows"]

            detail["params"] = [
                detail["params"],
                {
                    "type": "lineage",
                    "transform": self.name,
                    "dependencies": [
                        {"type": "dataframe", "nature": "input", "objects": [name]},
                        {
                            "type": "file",
                            "nature": "output",
                            "objects": [detail["filename"]],
                        },
                    ],
                },
            ]

            # Insert additional detail
            detail["transform"] = self.name
            detail["history"] = [{"transform": self.name, "log": "Wrote output"}]

            state.update_frame(name, detail)

        logger.debug(
            "Finished writing data",
            extra=self.config.get_extra({"transform": self.name}),
        )

    def validate_results(self, what, state):
        pass


provider = TableSink



import os
import logging
import click
from logging.config import dictConfig


def logging_start(filename=None):

    default_filename = "ext://sys.stderr"
    if filename is None:
        filename = default_filename
    elif isinstance(filename, str):
        try:
            os.makedirs(os.path.dirname(filename))
        except:

            pass
        filename = open(filename, "a")

    # logging.basicConfig(filename='example.log',level=logging.DEBUG)
    dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "basic": {
                    "class": "logging.Formatter",
                    "datefmt": "%I:%M:%S",
                    "format": "%(asctime)s %(levelname)s %(name)s %(message)s",
                },
                "json": {
                    "()": "pythonjsonlogger.jsonlogger.JsonFormatter",
                    "format": "%(asctime)s %(name)s %(levelname)s %(message)s",
                },
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "formatter": "basic",
                    "level": "DEBUG",
                    "stream": default_filename,
                },
                "app": {
                    "class": "logging.StreamHandler",
                    "formatter": "json",
                    "level": "DEBUG",
                    "stream": filename,
                },
            },
            "loggers": {
                "": {"level": "DEBUG", "handlers": ["console"], "propagate": False},
                "default": {
                    "level": "DEBUG",
                    "handlers": ["console"],
                    "propagate": False,
                },
                "app": {"level": "DEBUG", "handlers": ["app"], "propagate": False},
            },
            "root": {
                "level": "DEBUG",
                #'handlers': ['app', 'console'],
                "handlers": ["app"],
            },
        }
    )


def logging_end():
    logging.shutdown()


def logging_clear():
    for h in logging._handlers.copy():
        logging.removeHandler(h)
        h.flush()
        h.close()



import psutil
import resource
import logging
import platform

logger = logging.getLogger("app")


def set_limits_memory(params={}):

    # Default fraction of max memory to be used
    fraction = params.get("memory", 0.8)

    # Get available virtual memory
    vmem = psutil.virtual_memory()

    max_vmem = int(fraction * vmem.available)

    opsys = platform.system()
    release = int(platform.release().split('.')[0])

    # use AS as alternative to VMEM if the attribute isn't defined.
    # http://stackoverflow.com/a/30269998/5731870
    if hasattr(resource, "RLIMIT_VMEM"):
        resource.setrlimit(resource.RLIMIT_VMEM, (max_vmem, max_vmem))
    elif hasattr(resource, "RLIMIT_AS"):
        if opsys == 'Darwin' and release >= 21:
            # MacOS Monterey or higher
            # the setrlimit does not work
            pass
        else:
            resource.setrlimit(resource.RLIMIT_AS, (max_vmem, max_vmem))

    soft, hard = resource.getrlimit(resource.RLIMIT_AS)
    soft = round(soft / (1.0 * 1024 * 1024))
    hard = round(hard / (1.0 * 1024 * 1024))

    return "Memory: Soft ({}MB) Hard ({}MB)".format(soft, hard)


def set_limits(params={}):
    msg = set_limits_memory(params)

    logger.debug("Resource Limits set", extra={"transform": "Enricher", "data": msg})



import os, json, shutil, traceback
from collections import OrderedDict
from jinja2 import Environment, FileSystemLoader, PackageLoader, meta

from prompt_toolkit import prompt
from prompt_toolkit.token import Token
from prompt_toolkit.contrib.completers import PathCompleter
from prompt_toolkit.styles import style_from_dict
from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit.contrib.completers import WordCompleter

##########################
# Jinja2 helper functions...
##########################
def render(tpl_path, context):
    path, filename = os.path.split(tpl_path)
    return (
        Environment(loader=FileSystemLoader(path or "./"))
        .get_template(filename)
        .render(context)
    )


def get_variables(tpl_path):
    path, filename = os.path.split(tpl_path)
    env = Environment(loader=FileSystemLoader(path or "./"))
    template_source = env.loader.get_source(env, filename)[0]
    parsed_content = env.parse(template_source)
    return meta.find_undeclared_variables(parsed_content)


style = style_from_dict(
    {
        Token.Toolbar: "#ffffff bg:#333333",
    }
)

#################################################
# Helper
#################################################
def welcome():
    print("==============================")
    print("  Welcome to Scribble Enrich ")
    print("=============================")
    print("\n\n")


def read_globalconf():
    confpath = os.path.expanduser("~/.enrich.json")
    if os.path.exists(confpath):
        conf = json.loads(open(confpath).read())
    else:
        conf = {"workspaces": ["enrich-data"]}
    return conf


def write_globalconf(conf):
    confpath = os.path.expanduser("~/.enrich.json")
    with open(confpath, "w") as fd:
        fd.write(json.dumps(conf, indent=4))


#################################################
# Validators
#################################################
class NameValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0:
            raise ValidationError(
                message="Customer name should not be empty",
                cursor_position=len(document.text),
            )  # Move cursor to end of input.


class DescValidator(Validator):
    def validate(self, document):
        if len(document.text) == 0:
            raise ValidationError(
                message="Customer description should not be empty",
                cursor_position=len(document.text),
            )  # Move cursor to end of input.


class LogoValidator(Validator):
    def validate(self, document):
        url = document.text
        if len(url) == 0:
            raise ValidationError(
                message="Customer logo URL should not be empty",
                cursor_position=len(document.text),
            )  # Move cursor to end of input.
        if not url.startswith("http"):
            raise ValidationError(
                message="Customer logo URL should be a valid URL (http://...",
                cursor_position=len(document.text),
            )  # Move cursor to end of input.
        if not (
            url.lower().endswith("png")
            or url.lower().endswith("gif")
            or url.lower().endswith("jpg")
            or url.lower().endswith("ico")
        ):
            raise ValidationError(
                message="Customer logo URL should be a valid URL (http://....png",
                cursor_position=len(document.text),
            )  # Move cursor to end of input.


class ValidDir(Validator):
    def validate(self, document):
        if (
            (len(document.text) == 0)
            or (not os.path.exists(document.text))
            or (not os.path.isdir(document.text))
        ):
            raise ValidationError(
                message="Directory specified doesnt exist or is invalid",
                cursor_position=len(document.text),
            )


class ValidFile(Validator):
    def validate(self, document):
        if (
            (len(document.text) == 0)
            or (not os.path.exists(document.text))
            or (not os.path.isfile(document.text))
        ):
            raise ValidationError(
                message="File specified doesnt exist or is invalid",
                cursor_position=len(document.text),
            )


##########################
# Management
##########################
def create_workspace(workspace, context):

    # Create the directories
    try:
        os.makedirs(workspace)
    except:
        pass

    for subdir in [".", "conf", "output", "data", "transforms"]:
        try:
            os.makedirs(os.path.join(workspace, context["name"], subdir))
        except:
            pass

    # Now write the enrich.json
    thisdir = os.path.dirname(__file__)
    template = os.path.join(thisdir, "templates", "enrich.json.template")

    updated = render(template, context)
    enrichfile = os.path.join(workspace, context["name"], "enrich.json")
    with open(enrichfile, "w") as fd:
        fd.write(updated)


def prepare_workspace():

    welcome()

    print("We will ask you a few question\n")

    # => Directory
    path_completer = PathCompleter()
    workspace = prompt(
        "> Workspace directory: ",
        default="enrich-data",
        style=style,
        completer=path_completer,
    )

    print("\nWe need some additional input about the customer:\n")

    validators = {
        "name": NameValidator(),
        "description": DescValidator(),
        "logourl": LogoValidator(),
    }

    context = {}
    variables = get_variables(template)
    for v in variables:
        context[v] = prompt("> " + v + ": ", validator=validators.get(v, None))

    # => Update the global conf file...
    conf = read_globalconf()
    if workspace not in conf["workspaces"]:
        conf["workspaces"].append(workspace)
    write_globalconf(conf)


def prepare_transform():

    welcome()

    print("Bootstrapping a transform\n")

    conf = read_globalconf()
    workspaces = conf["workspaces"]

    # => Get workspace
    word_completer = WordCompleter(workspaces, ignore_case=True)
    if os.path.exists("enrich-data"):
        defaultdir = "enrich-data"
    else:
        defaultdir = "."

    workspace = prompt(
        "> Workspace: ",
        default=defaultdir,
        completer=word_completer,
        validator=ValidDir(),
    )

    # => Select the customer
    customers = os.listdir(workspace)
    if len(customers) == 0:
        print(
            "Looks like you havent created the workspace properly. Please run bootstrap workspace"
        )
        return

    customer_completer = WordCompleter(customers, ignore_case=True)
    customer = prompt(
        "> Customer: ", default=customers[0], completer=customer_completer
    )

    # => Bootstrap a template
    templatedir = os.path.join(os.path.dirname(__file__), "templates")

    # => Load the manifest
    print("\nPlease provide transform details\n")
    defaults = {"version": "1.0", "minorversion": "1.0"}
    validators = {"name": NameValidator(), "description": DescValidator()}
    manifest_template = os.path.join(templatedir, "manifest.json.template")
    context = {}
    variables = get_variables(manifest_template)
    for v in variables:
        context[v] = prompt(
            "> " + v + ": ",
            default=defaults.get(v, ""),
            validator=validators.get(v, None),
        )
    node_manifest = render(manifest_template, context)

    readme_template = os.path.join(templatedir, "README.md.template")
    node_readme = render(readme_template, context)

    # Get the right template
    templates = [t for t in os.listdir(templatedir) if t.endswith(".node.template")]
    templates = [t.replace(".node.template", "") for t in templates]
    template_completer = WordCompleter(templates, ignore_case=True)
    node = prompt(
        "> Type of Transform: ", default=templates[0], completer=template_completer
    )

    template_file = os.path.join(templatedir, "{}.node.template".format(node))
    node_content = render(template_file, context)

    # => Now you are ready to write
    transform_dir = os.path.join(workspace, customer, "transforms", context["name"])
    try:
        os.makedirs(transform_dir)
    except:
        pass

    with open(os.path.join(transform_dir, "__init__.py"), "w") as fd:
        fd.write(node_content)

    with open(os.path.join(transform_dir, "manifest.json"), "w") as fd:
        fd.write(node_manifest)

    with open(os.path.join(transform_dir, "README.md"), "w") as fd:
        fd.write(node_readme)

    # Bye bye
    print("Bootstrapped transform {} in {}".format(context["name"], transform_dir))


def prepare_config():

    welcome()

    print("Bootstrapping a new configuration\n")

    conf = read_globalconf()
    workspaces = conf["workspaces"]

    # => Get workspace
    word_completer = WordCompleter(workspaces, ignore_case=True)
    if os.path.exists("enrich-data"):
        defaultdir = "enrich-data"
    else:
        defaultdir = "."
    workspace = prompt(
        "> Workspace: ",
        default=defaultdir,
        completer=word_completer,
        validator=ValidDir(),
    )

    # => Select the customer
    usecases = os.listdir(workspace)
    if len(usecases) == 0:
        print(
            "Looks like you havent created the workspace properly. Please run bootstrap workspace"
        )
        return

    usecase_completer = WordCompleter(usecases, ignore_case=True)
    usecase = prompt("> Usecase: ", default=usecases[0], completer=usecase_completer)

    templatedir = os.path.join(os.path.dirname(__file__), "templates")
    validators = {"name": NameValidator(), "description": DescValidator()}

    print("\nPlease provide configuration details\n")
    config_template = os.path.join(templatedir, "config.json.template")
    context = {}
    variables = get_variables(config_template)
    for v in variables:
        context[v] = prompt("> " + v + ": ", validator=validators.get(v, None))
    config = render(config_template, context)

    name = context["name"]
    cleaned_name = "".join([x if x.isalnum() else "_" for x in name.lower()])
    configfile = os.path.join(
        workspace, customer, "conf", "{}.json".format(cleaned_name)
    )
    with open(configfile, "w") as fd:
        fd.write(config)

    # Bye bye
    print("Bootstrapped configuration {} in {}".format(context["name"], configfile))


def bootstrap(what):

    if what == "workspace":
        prepare_workspace()
    elif what == "transform":
        prepare_transform()
    elif what == "config":
        prepare_config()


def bootstrap1(config, transform, transformtype):
    """
    Initialize a new pipeline...
    """
    thisdir = os.path.dirname(__file__)

    # => First ensure that the destination exists
    if not os.path.exists(config):
        os.makedirs(config)
    elif os.path.isfile(config):
        raise Exception("Config path cannot be a file")

    # => Not install the transform
    if transform not in ["none", None]:

        transform_path = os.path.join(config, "transforms", transform)

        try:
            os.makedirs(transform_path)
        except:
            pass

        contentpath = os.path.join(thisdir, "{}.template".format(transformtype))
        if not os.path.exists(contentpath):
            raise Exception("Unsupported transform type template. Use generic")

        content = open(contentpath).read()
        initfile = os.path.join(transform_path, "__init__.py")
        with open(initfile, "w") as fd:
            fd.write(content)

        # Insert a manifest as well...
        manifest = os.path.join(transform_path, "manifest.json")
        with open(manifest, "w") as fd:
            fd.write(
                json.dumps(
                    {
                        "name": "MyModule",
                        "version": 1.0,
                        "minorversion": 1.0,
                        "params": [
                            {
                                "name": "Path",
                                "required": True,
                                "description": """\
Path for the module such as 
c:\path\to\file 
/path/to/file
""",
                            }
                        ],
                    },
                    indent=4,
                )
            )


############################################################
# Convert
############################################################
def convert(what, src, dst):
    """
    Take the old style transform skin and convert into a package each.
    """
    if os.path.exists(dst):
        raise Exception("Destination exists. Please remove it.")

    if not os.path.exists(src):
        raise Exception("Source does not exist. Invalid directory")

    basename = os.path.basename(src)
    try:
        enrichjson = os.path.abspath(os.path.join(src, "..", "..", "enrich.json"))
        usecase = json.load(open(enrichjson))
    except:
        traceback.print_exc()
        usecase = {"org": {"name": "Unknown", "customer": "Unknown"}}

    try:
        manifestcontent = json.load(open(os.path.join(src, "manifest.json")))
    except:
        manifestcontent = {}

    # =>
    defaults = {
        "what": what,
        "basename": basename,
        "name": basename,
        "version": "0.1",
        "description": basename + " package",
        "author": "Scribble Data",
        "email": "support@scribbledata.io",
        "url": "https://github.com/scribbledata/scribble-enrichsdk",
        "customer": usecase["org"]["name"],
        "usecase": usecase["org"]["name"],
    }

    params = {}
    for k, v in defaults.items():
        params[k] = manifestcontent.get(k, v)

    # Prepare the destination
    os.makedirs(dst)

    # Copy the code
    shutil.copytree(src, os.path.join(dst, basename))

    # => Write the manifest file
    # Write the manifest file...
    manifest = """\
include LICENSE
recursive-include %(basename)s *py *.html *md *template *json
"""
    manifest = manifest % params
    with open(os.path.join(dst, "MANIFEST.in"), "w") as fd:
        fd.write(manifest)

    # => Write license content
    license = """\
All rights reserved (c) Scribble Data          
"""
    with open(os.path.join(dst, "LICENSE"), "w") as fd:
        fd.write(license)

    # => Write the setup file
    setupcontent = """\
import os
from setuptools import setup

setup(name='%(basename)s',
      version='%(version)s',
      description='%(description)s',
      url='%(url)s', 
      author='%(author)s',
      author_email='%(email)s', 
      license='None',
      packages=['%(basename)s'],
      zip_safe=True,
      include_package_data=True,
      entry_points={
          '%(customer)s.%(what)s': ['%(basename)s=%(basename)s'],
      },
)
"""
    setupcontent = setupcontent % params

    with open(os.path.join(dst, "setup.py"), "w") as fd:
        fd.write(setupcontent)

    print("Created a package in {}".format(dst))



import json
import re
import logging

logger = logging.getLogger("app")

__all__ = ["BasePolicyEngine", "BaseResource", "BasePrincipal"]


class BaseResource(object):
    """Resource base class

    This will encapsulate data and matching patterns
    """

    def __init__(self, name, *args, **kwargs):
        if not isinstance(name, str) or len(name) == 0:
            raise Exception("Invalid resource name")

        self.name = name

    def match(self, name):
        return re.match(name, self.name) is not None

    def enforce(self, principal, action, args):
        return {"status": action, "details": args}


class BasePrincipal(object):
    """Principal base class

    This will encapsulate information about the principal
    (who) is asking for access
    """

    def __init__(self, identifiers, roles, attrs, *args, **kwargs):

        if (
            (identifiers is None)
            or (not isinstance(identifiers, list))
            or (len(identifiers) == 0)
        ):
            raise Exception("Invalid identifier specification")

        for i in identifiers:
            if (not isinstance(i, str)) or (len(i) == 0):
                raise Exception(
                    "Invalid identifier specification. Each identifier must be a non-trivial string"
                )

        if (roles is None) or (not isinstance(roles, dict)) or (len(roles) == 0):
            raise Exception("Invalid role specification")

        for r in roles:
            if (
                (not isinstance(r, str))
                or (len(r) == 0)
                or (not isinstance(roles[r], list))
                or (len(roles[r]) == 0)
            ):
                raise Exception(
                    "Invalid role specification Every role should be associated with a list of strings"
                )

        if (attrs is None) or (not isinstance(attrs, dict)) or (len(attrs) == 0):
            raise Exception("Invalid attribute specification. Not a dictionary")

        for a in attrs:
            if (not isinstance(a, str)) or (len(a) == 0):
                raise Exception(
                    "Invalid attribute specification. Every attribute name must be a string"
                )

        self.roles = roles
        self.identifiers = identifiers
        self.attrs = attrs

    def match(self, name):

        # First match the identifiers
        for i in self.identifiers:
            if re.match(name, i) is not None:
                return True

        for r, namelist in self.roles.items():
            for n in namelist:
                role = "Role::{}::{}".format(r, n)
                if re.match(name, role) is not None:
                    return True

        return False

    def get_attr(self, name):
        if not name.startswith("principal::"):
            raise Exception("Invalid attribute")

        if name not in self.attrs:
            raise Exception("Could not find attribute")

        return self.attrs[name]


class BasePolicyEngine(object):
    """Policy base class

    This is still work in progress. The goal is to support

    (a) row-level visibility of data (e.g., some store managers
        seeing only their store's data)
    (b) Possible anonymization and other special functions

    All principals and resources are strings that specify
    patterns. Examples include:

    (a) username::<name> (principal)
    (b) username::* (principal)
    (c) DailySummary::user_profile (resource)


    """

    NAME = "BasePolicy"

    def __init__(self, *args, **kwargs):

        self.spec = None
        self.args = None

    def configure(self, spec):
        """
        Configure an instance of the policy implementor with a
        specification

        Args:
          spec (dict): Specification

        """

        if not isinstance(spec, dict):
            raise Exception("Invalid specification format")

        self.spec = spec
        self.args = spec.get("args", {})

    def validate(self):
        """
        Validate the specification. Raise an exception if
        invalid
        """

        if self.args is None:
            raise Exception("Policy not configured")

        if "policies" not in self.args:
            raise Exception("Missing policy specification")

        policies = self.args["policies"]

        if not isinstance(policies, list):
            raise Exception("Policy specification must be a list")

        if len(policies) == 0:
            raise Exception("Empty list of policies")

        for p in policies:

            # => Is the specification valid?
            if (
                (not isinstance(p, dict))
                or (("principal" not in p) or (p["principal"] in [None, ""]))
                or (("resource" not in p) or (p["resource"] in [None, ""]))
                or (("action" not in p) or (p["action"] in [None, ""]))
            ):
                raise Exception(
                    "Invalid specification. Each policy should be a dictionary with a definition of principal, resource, and actions "
                )

            action = p["action"]
            if action not in ["allow", "deny", "filter"]:
                raise Exception("Invalid action in specification")

            if "args" in p and not isinstance(p["args"], dict):
                raise Exception(
                    "Invalid specification. Action args when specified should be a dict"
                )

            nomatch = p.get("nomatch", "pass")
            if nomatch not in ["allow", "deny", "pass"]:
                raise Exception(
                    "Invalid specification. Nomatch rule must be either allow, deny, or pass. Default is pass"
                )

    def apply(self, principal, resource):
        """
        Apply the policy specification

        Args:
          principal (str): username/other identifier
          resource (dict): resource name

        Returns:
          dict: Output with result of the policy check ('status') and the data

        """

        deny_response = {"status": "deny"}

        if not issubclass(principal.__class__, BasePrincipal):
            raise Exception("Principal should be a subclass of BasePrincipal")

        if not issubclass(resource.__class__, BaseResource):
            raise Exception("resource should be a subclass of BaseResource")

        for i, p in enumerate(self.args["policies"]):

            name = p.get("name", "Policy{}".format(i))
            logger.debug("Matching:", name)
            logger.debug("Principal:", p["principal"], principal.match(p["principal"]))
            logger.debug("Resource:", p["resource"], resource.match(p["resource"]))

            nomatch = p.get("nomatch", "pass")
            args = p.get("args", {})

            if principal.match(p["principal"]) and resource.match(p["resource"]):
                action = p["action"]
                logger.debug("Action:", action)
                if action == "deny":
                    return deny_response
                return resource.enforce(principal, action, args)
            else:
                if nomatch == "pass":
                    continue
                elif nomatch == "allow":
                    return resource.enforce(principal, nomatch, args)
                else:
                    return deny_response

        # No match for any policy
        return {"status": "deny"}


if __name__ == "__main__":
    print(vars())
    subclasses = vars()["BasePolicy"].__subclasses__()
    for cls in subclasses:
        print(cls.NAME)



import re
import os
import sys
import py_compile
import ast
from setuptools import setup, find_packages

_version_re = re.compile(r"VERSION\s+=\s+(.*)")

thisdir = os.path.dirname(__file__)
readme = open(os.path.join(thisdir, "README.rst")).read()

with open("enrichsdk/__init__.py", "rb") as f:
    version = str(
        ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))
    )

setup(
    name="enrichsdk",
    version="3.7.6",
    description="Enrich Developer Kit",
    long_description=readme,
    url="http://github.com/pingali/scribble-enrichsdk",
    author="Venkata Pingali",
    author_email="pingali@scribbledata.io",
    license="All rights reserved",
    scripts=[],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "wheel",
        "docutils<0.16",
        "click==7.1.2",
        "aiobotocore==1.2.1",
        "aioitertools==0.8.0",
        "boto3==1.16.4",
        "botocore==1.19.52",
        "sphinx-click>=2.3.0",
        "glob2==0.7",
        "requests>=2.21.0,<2.27",
        "requests-oauthlib==0.8.0",
        "pytest>=4.6.0",
        "numpydoc>=0.7.0",
        "pandas==0.25.3",
        "dask==2.30.0",
        "distributed==2.30.1",
        "idna==2.8",
        "coverage==5.5",
        "flake8==3.9.0",
        "raven==6.6.0",
        "python-json-logger==0.1.8",
        "python-dateutil==2.8.1",
        "s3fs==0.5.1",
        "fsspec==0.8.5",
        "colored==1.3.5",
        "flask-multistatic==1.0",
        "humanize==0.5.1",
        "pytz==2020.1",
        "Flask==2.0.3",
        "Jinja2==3.0.3",
        "pytest-cov",
        "Markdown>=2.9.10",
        "prompt-toolkit>=2.0.10",
        "pyarrow>=0.9.0",
        "cytoolz>=0.9.0.1",
        "jsonschema>=3.2.0",
        "flask_cors",
        #'moto>=1.3.14',
        "prefect==0.15.7",
        "distro>=1.4.0",
        "gcsfs==0.6.0",
        "jupyter-core>=4.6.1",
        "nbformat>=5.1.2",
        "tzlocal>=2.0.0",
        "texttable",
        "pykafka",
        "redis",
        "gitpython",
        "logstash_formatter",
        "pyhive",
        "pyfiglet",
        "sqlalchemy>=1.4.0",
        "kafka-python==2.0.2",
        "pykafka==2.8.0",
        "papermill>=2.3.0",
        "sqllineage",
    ],
    entry_points={
        "console_scripts": [
            "enrichpkg=enrichsdk.scripts.enrichpkg:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import os
import sys
import io
import json
import copy
import logging
import string
import random
import csv
import gzip
import hashlib
import math
import traceback
import tempfile
import gc
import sqlite3
from collections import OrderedDict, defaultdict
from datetime import datetime, timedelta
from dateutil import parser as dateparser
import pandas as pd
import numpy as np

from django.shortcuts import render
from django.http import HttpResponse, JsonResponse
from django.utils.http import urlencode
from django.views.decorators.csrf import csrf_exempt
from django.conf import settings
from django.urls import reverse
from django.db import models
from django.http import HttpResponseRedirect, Http404
from django.contrib import messages
from django.core.mail import send_mail, BadHeaderError
from django.core.mail import EmailMultiAlternatives
import enrichsdk

from celery.result import AsyncResult
from celery import current_app

from users.decorators import log_activity
from enrichsdk.lib.customer import find_usecase

from .tasks import complex_task

logger = logging.getLogger("app")

usecase = find_usecase(__file__)


def index(request):
    return render(
        request,
        "APPNAME/index.html",
        {
            "usecase": usecase,
        },
    )


@log_activity("complex", nature="application")
def complex_request(request):
    """
    Use this for workflows.
    """

    post = request.POST.dict()
    params = post["params"]
    referrer = post["referrer"]
    try:

        # Gather the parameters
        params = json.loads(params)
        name = params.get("name", "")
        dt = datetime.now().date().isoformat()
        name = f"complex-{name}-{dt}"
        # Now look for transactions
        result = complex_task.delay(
            {
                "name": name,
                "source": params.get("source", ""),
            }
        )

    except:
        logger.exception("Internal error")
        messages.error(request, "Error while searching. Check log")
        return HttpResponseRedirect(referrer)

    try:
        resulturl = (
            reverse("APPNAME:txnsearch_result")
            + "?"
            + urlencode({"task_id": result.id, "referrer": referrer})
        )

        return HttpResponseRedirect(resulturl)
    except:
        logger.exception("Unable to issue a background task")
        messages.error(request, "Unable to issue a background task")
        return HttpResponseRedirect(referrer)


def complex_result(request):
    """
    Show the results...
    """

    task_id = request.GET.get("task_id", None)
    referrer = request.GET.get("referrer", None)
    download = request.GET.get("download", None)
    download = str(download) == "true"

    # Handle invalid referrer...
    if referrer is None:
        messages.warning(request, f"No referrer")
        referrer = reverse("APPNAME:index")

    # Sanity check
    if task_id is None:
        messages.error(request, f"Empty task id")
        return HttpResponseRedirect(referrer)

    thisurl = (
        reverse("APPNAME:complex_result")
        + "?"
        + urlencode({"task_id": task_id, "referrer": referrer})
    )

    try:

        task_result = AsyncResult(task_id)

        value = task_result.result
        if not isinstance(value, dict):
            value = str(value)

        result = {
            "id": task_id,
            "status": task_result.status,
            "result": value,
            "args": task_result.args,
        }

    except:
        logger.exception("Unable to access result")
        messages.warning(request, f"Internal error. Invalid task")
        return HttpResponseRedirect(referrer)

    if download:

        if result["status"] != "SUCCESS":
            messages.error(request, f"Download only supported for successful runs")
            return HttpResponseRedirect(thisurl)

        # Now good to go
        records = result["result"]["records"]
        if len(records) == 0:
            messages.error(request, f"No matches found")
            return HttpResponseRedirect(referrer)

        name = result["args"][0].get("name", "complex-task-{task_id}.csv")
        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = f'attachment; filename="{name}"'

        # Write the content...
        writer = csv.writer(response)
        for r in records:
            writer.writerow(r)

        # This renders the csv...
        return response

    return render(
        request,
        "APPNAME/complex_task_result.html",
        {
            "usecase": usecase,
            "task": result,
            "referrer": referrer,
        },
    )



import os
import sys
import json

from .schema import *


def post(backend, service=None, filename=None, data=None, debug=False):
    """
    Post featurestore to server

    Args:
      backend (object): Backend class enrichsdk.api.Backend
      service (object): Dictionary with name, path
      filename (str): Path to the input file to the posted
      data (dict): Dictionary to be posted (if filename not specified)

    Returns
      dict: Response dictionary from the server

    """

    if (
        (service is None)
        or (not isinstance(service, dict))
        or ("name" not in service)
        or (service["name"] not in ["run", "spec"])
    ):
        raise Exception("Unsupported service")

    url = service["path"] + "/add/?format=json"
    if not url.startswith("http"):
        url = backend.get_api_url(url, use_prefix=False)

    if debug:
        print("Backend URL: {}".format(url))

    if filename is None and data is None:
        raise Exception("Post requires either filename or data")

    try:
        if data is None:
            data = json.load(open(filename))
    except:
        raise Exception("Require a valid json file to post to server")

    name = service["name"]
    if name == "run":
        obj = FeatureGroupRun()
    elif name == "spec":
        obj = FeatureGroupSpec()
    else:
        raise Exception("Unknown service name: {}".format(name))

    # Add attributes
    for k, v in data.items():
        obj.add(k, v)

    # Check if the schema is valid
    obj.validate()

    data = obj.export()
    if debug:
        print("Data")
        print(json.dumps(data, indent=4))

    response = backend.call(url, method="post", data=data)
    return response


def download(
    backend, service=None, featuregroup_id=None, run_id=None, data=None, debug=False
):
    """
    Post featurestore to server

    Args:
      backend (object): Backend class enrichsdk.api.Backend
      service (object): Dictionary with name, path
      featuregroup_id (int): Id of the featuregroup to download
      run_id (int): Id of the run to download
      data (dict): Dictionary to be posted (if filename not specified)

    Returns
      dict: Response dictionary from the server

    """

    if (
        (service is None)
        or (not isinstance(service, dict))
        or ("name" not in service)
        or (service["name"] not in ["run", "spec"])
    ):
        raise Exception("Unsupported service")

    if featuregroup_id is not None:
        suffix = "/download/{}/?format=json".format(featuregroup_id)
    elif run_id is not None:
        suffix = "/download/{}/?format=json".format(run_id)
    else:
        raise Exception("One of featuregroup_id or run_id should be specified")

    url = service["path"] + suffix

    if not url.startswith("http"):
        url = backend.get_api_url(url, use_prefix=False)

    if debug:
        print("Backend URL: {}".format(url))

    response = backend.call(url, method="get")

    return response


def generate(backend, service=None, debug=False):
    """
    Generate sample specification files
    """

    if (
        (service is None)
        or (not isinstance(service, dict))
        or ("name" not in service)
        or (service["name"] not in ["run", "spec"])
    ):
        raise Exception("Unsupported service")

    name = service["name"]
    if name == "run":
        obj = FeatureGroupRun()
    elif name == "spec":
        obj = FeatureGroupSpec()
    else:
        raise Exception("Unknown service name: {}".format(name))

    return obj.export(dummy=True)


def search(backend, service=None, debug=False, params={}):
    """
    Search featurestore for specific featuregroups

    Args:
      backend (object): Backend class enrichsdk.api.Backend
      service (object): Dictionary with name, path
      args (dict): Search criteria as key, value paits
      debug (bool): Debug run or not

    Returns
      dict: Response dictionary from the server

    """

    if (
        (service is None)
        or (not isinstance(service, dict))
        or ("name" not in service)
        or (service["name"] not in ["run", "spec"])
    ):
        raise Exception("Unsupported service")

    suffix = "/search/?format=json"
    url = service["path"] + suffix

    if not url.startswith("http"):
        url = backend.get_api_url(url, use_prefix=False)

    if debug:
        print("Backend URL: {}".format(url))

    response = backend.call(url, method="get", params=params)

    return response
